from datetime import date, datetime, timedelta
import io
import json
import logging
import os
import re
import inspect

import boto3
from botocore import UNSIGNED
from botocore.config import Config
from botocore.client import ClientError
import numpy as np
from obspy.clients.filesystem.sds import Client as SDS_Client
from obspy.clients.fdsn import Client as FDSN_Client
from obspy.clients.fdsn.header import FDSNException

from obspy import (read_inventory,
                   Inventory,
                   Trace,
                   Stream,
                   UTCDateTime,
                   read)
import pandas as pd
import xarray as xr
import yaml

from zizou.util import test_signal


logger = logging.getLogger(__name__)


class PostProcessException(Exception):
    pass

class PostProcess:
    """
    Class for checking waveforms and station metadata for valid values,
    filling waveform data gaps and removing instrument sensitivity
    """
    def __init__(self, st=Stream(), inv=Inventory(), inv_dt=Inventory(),
                 startdate=None, enddate=None, loc='*', comp='*Z',
                 fill_value=np.nan):
        self.st = st
        self.inv = inv
        self.inv_dt = inv_dt
        self.startdate = startdate
        self.enddate = enddate
        self.loc = loc
        self.comp = comp
        self.fill_value = fill_value

        # Get check registry
        self.checklist = [
            x for x in inspect.getmembers(self)
            if inspect.ismethod(x[1]) and x[0].startswith("check")
        ]
        # Get matching post-processing functions
        self.pp_functions = []
        for check in self.checklist:
            self.pp_functions.append(
                [x for x in inspect.getmembers(self)
                 if x[0] == check[0].replace("check", "output")
                 ][0]
            )

        # Run checks
        self.res = None

    def run_post_processing(self):
        self.res = self.run_checks()
        func = [a[1] for (a, b) in zip(self.pp_functions, self.res) if b][0]
        trace = func()
        return trace

    def run_checks(self):
        self.res = [check[1]() for check in self.checklist]
        if sum(self.res) == 0:
            raise NotImplementedError(
                'No post processing not set up for this case:\n'
                '{}'.format(self.st)
            )
        elif sum(self.res) > 1:
            msg = 'More than one check is true: {}. Processing workflow ' \
                  'is unclear.'.format(', '.join(
                [a[0] for (a, b) in zip(self.checklist, self.res) if b])
            )
            raise PostProcessException(msg)
        else:
            return self.res

    def check_case_no_data(self):
        """
        Checks if either no stream, or no station metadata exist.
        """
        if len(self.st) < 1 or len(self.inv_dt) < 1:
            return True
        elif len(self.inv_dt[0][0]) == 1 and \
                self.inv_dt[0][0][0].start_date == self.enddate:
            return True
        elif len(self.inv_dt[0][0]) == 1 and \
                self.inv_dt[0][0][0].end_date == self.startdate:
            return True
        else:
            return False

    def output_case_no_data(self):
        raise  PostProcessException

    def check_case_multiple_channel_periods(self):
        """
        Checks if there are multiple channel operation periods in a given
        time period. If so, different response removal is required.
        """
        if len(self.inv_dt) < 1 or len(self.st) < 1:
            return False
        elif len(self.inv_dt[0][0]) > 1:
            return True
        else:
            return False

    def output_case_multiple_channel_periods(self):
        """
        Stitches two response periods together to make a single merged
        trace.
        """
        if len(self.inv_dt[0][0]) > 2:
            raise NotImplementedError(
                'Current post processing function only valid for a '
                'maximum of two response periods.'
            )
        st1 = self.st.copy()
        st1.merge(fill_value=self.fill_value)
        st2 = st1.copy()
        st1.trim(starttime=self.startdate, endtime=self.inv_dt[0][0][0].end_date,
                 nearest_sample=False)
        st2.trim(starttime=self.inv_dt[0][0][1].start_date, endtime=self.enddate,
                 nearest_sample=False)
        st1.attach_response(self.inv_dt)
        st1.remove_sensitivity()
        st2.attach_response(self.inv_dt)
        st2.remove_sensitivity()
        st1 += st2
        st1.merge(fill_value=self.fill_value)
        st1.trim(
            self.startdate, self.enddate, nearest_sample=False,
            pad=True, fill_value=self.fill_value
        )
        return st1[0]

    def check_case_incomplete_station_metadata(self):
        """
        The case where a a channel period start/stops during the
        selected time window. Some parts of the trace in the stream
        might not correspond to a valid period according to station
        metadata.
        """
        if len(self.inv_dt) < 1:
            return False
        elif len(self.inv_dt[0][0]) != 1 or len(self.st) < 1:
            return False
        elif ((self.startdate < self.inv_dt[0][0][0].start_date < self.enddate) or
                (self.startdate < self.inv_dt[0][0][0].end_date < self.enddate)):
            return True
        else:
            return False

    def output_case_incomplete_station_metadata(self):
        if self.inv_dt[0][0][0].start_date > self.startdate:
            self.st.trim(self.inv_dt[0][0][0].start_date, self.enddate,
                         nearest_sample=False)
            self.st.attach_response(self.inv_dt)
            self.st.remove_sensitivity()
            self.st.trim(
                self.startdate, self.enddate, nearest_sample=False,
                pad=True, fill_value=self.fill_value
            )
            return self.st[0]
        elif self.inv_dt[0][0][0].end_date < self.enddate:
            self.st.trim(self.startdate, self.inv_dt[0][0][0].end_date,
                         nearest_sample=False)
            self.st.attach_response(self.inv_dt)
            self.st.remove_sensitivity()
            self.st.trim(
                self.startdate, self.enddate, nearest_sample=False,
                pad=True, fill_value=self.fill_value
            )
            return self.st[0]
        else:
            raise PostProcessException('Post-procesing function not equipped to'
                            'handle:\n'
                            '{}'.format(self.st)
                            )

    def check_case_no_station_issues(self):
        if len(self.inv_dt) < 1:
            return False
        elif len(self.inv_dt[0][0]) != 1 or len(self.st) < 1:
            return False
        elif len(self.inv_dt[0][0]) == 1 and \
                self.inv_dt[0][0][0].end_date == self.startdate:
            return False
        elif len(self.inv_dt[0][0]) == 1 and \
                self.inv_dt[0][0][0].start_date == self.enddate:
            return False
        elif (self.startdate < self.inv_dt[0][0][0].start_date < self.enddate) \
                or (self.startdate < self.inv_dt[0][0][0].end_date < self.enddate):
            return False
        else:
            return True

    def output_case_no_station_issues(self):
        self.st.merge(fill_value=self.fill_value)
        self.st.trim(
            starttime=self.startdate, endtime=self.enddate,
            nearest_sample=False, pad=True, fill_value=self.fill_value
        )
        self.st.attach_response(self.inv_dt)
        self.st.remove_sensitivity()
        return self.st[0]


def get_instrument_response(net, site, loc,
                            staxml_dir='./instrument_response',
                            fdsn_urls=('https://service.geonet.org.nz')):
    """
    Retrieve instrument response file in STATIONXML format.
    """
    fn = '{:s}.xml'.format(site)
    fn = os.path.join(staxml_dir, fn)
    try:
        inv = read_inventory(fn, format='STATIONXML')
    except FileNotFoundError:
        try:
            os.makedirs(staxml_dir)
        except FileExistsError:
            pass
        for url in fdsn_urls:
            try:
                client = FDSN_Client(base_url=url)
                inv = client.get_stations(network=net,
                                          location=loc,
                                          station=site,
                                          channel='*',
                                          level='response')
            except Exception as e:
                logger.info(e)
            else:
                break
        inv.write(fn, format='STATIONXML')
    return inv


class SDSWaveforms(object):

    def __init__(self, sds_dir, fdsn_urls, staxml_dir,
                 fill_value=np.nan):
        """
        Get seismic waveforms from a local SDS archive.
        """
        self.client = SDS_Client(sds_dir)
        self.fdsn_urls = fdsn_urls
        self.staxml_dir = staxml_dir
        self.fill_value = fill_value

    def get_waveforms(self, net, site, loc, comp, startdate, enddate):
        """
        :param fill_value: Default value used to fill gaps and pad traces to
                           cover the whole time period defined by `startdate`
                           and `enddate`. If 'interpolate' is chosen, the
                           traces will not be padded.
        :type fill_value: int, float, 'interpolate', np.nan, or None
        """
        inv = get_instrument_response(net, site, loc,
                                      staxml_dir=self.staxml_dir,
                                      fdsn_urls=self.fdsn_urls)

        st = self.client.get_waveforms(net, site, loc, comp, startdate,
                                       enddate, dtype='float64')

        inv_dt = inv.select(location=loc, channel=comp,
                            starttime=startdate,
                            endtime=enddate - 1)

        # Post-process to remove sensitivity and account for gaps
        pp = PostProcess(st=st, inv=inv, inv_dt=inv_dt,
                         startdate=startdate, enddate=enddate,
                         loc=loc, comp=comp, fill_value=self.fill_value)
        tr = pp.run_post_processing()
        return tr


class MockSDSWaveforms(object):
    """
    Mock SDSWaveforms class for testing by creating
    synthetic data for the requested streams.
    """

    def __init__(self, sds_dir):
        """
        Get seismic waveforms from a local SDS archive.
        """
        os.makedirs(sds_dir, exist_ok=True)
        self.client = SDS_Client(sds_dir)

    def save2sds(self, trace: Trace, rootdir: str):
        """
        Save a trace to a SDS directory structure.

        Parameters
        ----------
        trace : `obspy.Trace` 
            Seismic/acoustic trace to be written to disk. 
        rootdir : str 
            Root directory for the SDS directory structure. 
        """
        sds_fmtstr = os.path.join(
            "{year}", "{network}", "{station}", "{channel}.{sds_type}",
            "{network}.{station}.{location}.{channel}.{sds_type}.{year}.{doy:03d}")

        fullpath = sds_fmtstr.format(year=trace.stats.starttime.year,
                                    doy=trace.stats.starttime.julday,
                                    sds_type="D", **trace.stats)
        fullpath = os.path.join(rootdir, fullpath)
        dirname, filename = os.path.split(fullpath)
        os.makedirs(dirname, exist_ok=True)
        print("writing", fullpath)
        trace.write(fullpath, format="mseed")


    def generate_dataset(self, rootdir: str, net: str, site: str,
                         loc: str, comp: str, start: str, end: str) -> str:
        """ 
        Generate a test dataset for the integration tests.

        Parameters
        ----------
        rootdir : str
            Parent directory for the test dataset.
        start : str
            Start time for the test dataset in ISO 8601 format.
        end : str
            End time for the test dataset in ISO 8601 format.
        """ 
        tstart = UTCDateTime(start)
        # Always generate whole day files
        _tstart = UTCDateTime(year=tstart.year, julday=tstart.julday)
        tend = UTCDateTime(end)
        tr = test_signal(starttime=_tstart,
                        sampling_rate=10,
                        nsec=86400, gaps=True,
                        network=net, station=site,  
                        location=loc, channel=comp)
        while _tstart < tend:
            tr.stats.starttime = _tstart
            self.save2sds(tr, rootdir)
            _tstart += 86400


    def get_waveforms(self, net, site, loc, comp, startdate, enddate):
        """
        Generate synthetic daily files that start at midnight on the
        startdate and end at midnight on the enddate. The return the requested
        timespan from these files.
        """
        self.generate_dataset(self.client.sds_root, net, site, loc, comp,
                              startdate, enddate)
        st = self.client.get_waveforms(net, site, loc, comp, startdate,
                                       enddate, dtype='float64')
        return st[0]


class FDSNWaveforms(object):

    def __init__(self, url, debug=False, fill_value=np.nan):
        """
        Get seismic waveforms from FDSN web service.
        """
        self.client = FDSN_Client(base_url=url, debug=debug)
        self.fill_value = fill_value

    def get_waveforms(self, net, site, loc, comp, startdate, enddate):
        st = self.client.get_waveforms(net, site, loc, comp, startdate,
                                       enddate, attach_response=True)
        # Prepare seismic time series
        st.remove_sensitivity()
        # in case stream has more than one trace
        st.merge(fill_value=self.fill_value)
        if self.fill_value != 'interpolate':
            st.trim(startdate, enddate, pad=True, fill_value=self.fill_value,
                    nearest_sample=False)
        return st[0]


class S3Waveforms(object):

    def __init__(self, s3bucket, fdsn_urls, staxml_dir, debug=False,
                 fill_value=np.nan):
        config = Config(signature_version=UNSIGNED)
        self.client = boto3.client('s3', config=config)
        self.s3_bucket = s3bucket 
        self.fdsn_urls = fdsn_urls
        self.staxml_dir = staxml_dir
        self.debug = debug
        self.fill_value = fill_value


    def get_waveforms(self, net, site, loc, comp, start, end):
        """
        Get seismic waveforms from GeoNet's open data archive on AWS S3.
        """
        PATH_FORMAT = "waveforms/miniseed/{year}/{year}.{julday:03d}/{station}."
        PATH_FORMAT += "{network}/{year}.{julday:03d}.{station}."
        PATH_FORMAT += "{location}-{channel}.{network}.D"
        t_start = start
        t_end = UTCDateTime(year=t_start.year,
                            julday=t_start.julday + 1,
                            hour=0, minute=0,
                            second=0, microsecond=0)
        t_end = min(t_end, end)
        st = Stream()
        while t_start < end:
            s3_file_path = PATH_FORMAT.format(year=t_start.year,
                                              julday=t_start.julday,
                                              station=site,
                                              network=net,
                                              location=loc,
                                              channel=comp)
            data = io.BytesIO()
            if self.debug:
                logger.debug('Requesting {}'.format(s3_file_path))
            _ = self.client.download_fileobj(Bucket=self.s3_bucket,
                                             Key=s3_file_path,
                                             Fileobj=data)
            data.seek(0)
            _st = read(data, dtype='float64')
            data.close()
            _st.trim(t_start, t_end, nearest_sample=False)
            st += _st
            t_start = t_end
            t_end = t_start + timedelta(days=1)
            t_end = min(t_end, end)

        st.merge(fill_value=self.fill_value)
        inv = get_instrument_response(net, site, loc,
                                      staxml_dir=self.staxml_dir,
                                      fdsn_urls=self.fdsn_urls)
        inv_dt = inv.select(location=loc, channel=comp,
                            starttime=t_start,
                            endtime=t_end)

        # Post-process to remove sensitivity and account for gaps
        pp = PostProcess(st=st, inv=inv, inv_dt=inv_dt,
                         startdate=start, enddate=end,
                         loc=loc, comp=comp, fill_value=self.fill_value)
        tr = pp.run_post_processing()
        return tr
        

class DataSource:

    def __init__(self, sds_dir='/geonet/seismic',
                 staxml_dir='/scratch/zizou/STATIONXML', 
                 s3bucket='geonet-open-data',
                 fdsn_urls=('https://service.geonet.org.nz',
                            'https://service-nrt.geonet.org.nz'),
                 debug=False, chunk_size=None,
                 sources = ('sds', 'fdsn', 's3'),
                 fill_value=np.nan):
        
        if len(sources) < 1:
            msg = "Sources can't be empty!"
            raise ValueError(msg)
        
        for src in sources:
            if src not in ('sds', 'fdsn', 's3', 'mock'):
                msg = "Sources have to be one or several of 'sds', 'fdsn', 's3' 'mock': " + src
                raise ValueError(msg)
        self.staxml_dir = staxml_dir
        self.chunk_size = chunk_size
        self.clients = []
        if 'sds' in sources:
            try:
                self.clients.append(SDSWaveforms(sds_dir=sds_dir,
                                                 fdsn_urls=fdsn_urls,
                                                 staxml_dir=staxml_dir,
                                                 fill_value=fill_value))
            except OSError as e:
                logger.error(e)    

        if 's3' in sources:
            try:
                self.clients.append(S3Waveforms(s3bucket=s3bucket,
                                                fdsn_urls=fdsn_urls,
                                                staxml_dir=staxml_dir,
                                                fill_value=fill_value))
            except ClientError as e:
                logger.error(e)
            
        if 'fdsn' in sources:
            for url in fdsn_urls: 
                try:
                    logger.info("Adding FDSN client for {}".format(url))
                    self.clients.append(FDSNWaveforms(url=url, debug=debug,
                                                      fill_value=fill_value))
                except FDSNException as e:
                    logger.error(e)        
        if 'mock' in sources:
            self.clients.append(MockSDSWaveforms(sds_dir=sds_dir))

    def get_waveforms(self, net, site, loc, comp, start, end):
        t_start = start
        t_end = end 
        if self.chunk_size is not None:
            t_end = t_start + self.chunk_size
            t_end = min(t_end, end)
        while t_end <= end:
            tr = Trace()
            for client in self.clients:
                try:
                    tr = client.get_waveforms(net, site, loc, comp, t_start, t_end)
                except Exception as e:
                    logger.info(e)
                    continue
                else:
                    break
            if not tr:
                msg = "No data found for {}"
                logger.error(msg.format('.'.join((net, site, loc, comp))))
            
            yield tr

            if self.chunk_size is not None:
                t_start = t_end 
                cond1 = (end - t_end) < self.chunk_size
                cond2 = end > t_end 
                if cond1 and cond2:
                    last_chunk = end - t_end
                    t_end = t_start + last_chunk
                else:
                    t_end = t_start + self.chunk_size
            else:
                # go beyond the end to stop the loop
                t_end = end + 1.


def hashForFeatureRequest__Call(args, kwargs):
    key = []
    self = args[0]
    key.append((self.volcano, self.site, self.channel,
                str(self.starttime), str(self.endtime)))
    for _a in args[1:]:
        key.append(_a)
    key = tuple(key)    
    key += tuple(sorted(kwargs.items()))
    hashValue = hash(key)
    return hashValue


class FeatureRequest:
    """
    Query computed features

    :param rootdir: Path to parent directory.
    :type rootdir: str
    :param starttime: Begin of request
    :type starttime: :class:`datetime.datetime`
    :param endtime: Begin of request
    :type endtime: :class:`datetime.datetime`

    >>> import datetime
    >>> fq = FeatureRequest()
    >>> start = datetime.datetime(2012,1,1,0,0,0)
    >>> end = datetime.datetime(2012,1,2,23,59,59)
    >>> volcano = 'Whakaari'
    >>> site = 'WIZ'
    >>> chan = 'HHZ'
    >>> fq.volcano = volcano
    >>> fq.starttime = start
    >>> fq.endtime = end
    >>> fq.site = site
    >>> fq.channel = chan
    >>> rsam = fq("rsam")
    """
    features1D = ('dsar', 'rsam', 'rsam_energy_prop',
                  'central_freq', 'bandwidth', 'predom_freq')
    features2D = ('ssam', 'sonogram', 'filterbank')
    mlresults2D = ('autoencoder',)
    mlresults1D = ('autoencoder_indicator', 'decision_tree', 'random_walk')
    features = features1D + features2D + mlresults2D + mlresults1D
    feat_dict = {'1D': features1D, '2D': features2D, 'ml2D': mlresults2D,
                 'ml1D': mlresults1D, 'all': features}

    def __init__(self, rootdir='/scratch/zizou/features',
                 volcano=None, site=None, channel=None,
                 interval='10min', 
                 starttime=datetime.utcnow()-timedelta(seconds=600),
                 endtime=datetime.utcnow()):

        self.volcano = volcano
        self.site = site
        self.channel = channel
        self.starttime = starttime
        self.endtime = endtime
        self.rootdir = rootdir
        self.interval = interval

    def __call__(self, feature, stack_length=None):
        """
        Request a particular feature

        :param feature: Feature name
        :type feature: str
        :param stack_length: length of moving average in time
        :type stack_length: str

        """
        if self.endtime <= self.starttime:
            raise ValueError('Startime has to be smaller than endtime.')

        logger.debug(f"Reading feature {feature} between {self.starttime} and {self.endtime}")
        feature = feature.lower()
        num_periods = None
        if stack_length is not None:
            valid_stack_units = ['W', 'D', 'H', 'T', 'min', 'S']
            if not re.match(r'\d*\s*(\w*)', stack_length).group(1)\
                   in valid_stack_units:
                raise ValueError(
                    'Stack length should be one of: {}'.
                        format(', '.join(valid_stack_units))
                )

            if pd.to_timedelta(stack_length) < pd.to_timedelta(self.interval):
                raise ValueError('Stack length {} is less than interval {}'.
                                 format(stack_length, self.interval))

            # Rewind starttime to account for stack length
            self.starttime -= pd.to_timedelta(stack_length)

            num_periods = (pd.to_timedelta(stack_length)/
                           pd.to_timedelta(self.interval))
            if not num_periods.is_integer():
                raise ValueError(
                    'Stack length {} / interval {} = {}, but it needs'
                    ' to be a whole number'.
                        format(stack_length, self.interval, num_periods))

        if feature not in self.features:
            raise ValueError('Feature {} does not exist.'.format(feature))

        xd_index = dict(datetime=slice(self.starttime,
                                       (self.endtime-
                                        pd.to_timedelta(self.interval))))
        filename = os.path.join(self.sitedir,
                                '%s.nc' % feature)
        with xr.open_dataset(filename, group='original', engine='h5netcdf') as ds:
            ds.sortby("datetime")
            rq = ds.loc[xd_index].load()

        # Stack features

        if stack_length is not None:
            logger.debug("Stacking feature...")
            try:
                xdf = rq[feature].rolling(datetime=int(num_periods),
                                        center=False,
                                        min_periods=1).mean()
                # Return requested timeframe to that defined in initialisation
                self.starttime += pd.to_timedelta(stack_length)
                xdf_new = xdf.loc[
                        self.starttime:
                        self.endtime-pd.to_timedelta(self.interval)]
                xdf_new = xdf_new.rename(feature)
            except ValueError as e:
                logger.error(e)
                logger.error('Stack length {} is not valid for feature {}'.
                             format(stack_length, feature))
            else:
                return xdf_new

        return rq[feature]

    def get_starttime(self):
        return self.__starttime

    def set_starttime(self, time):
        if time is None:
            self.__starttime = None
            self.__sdate = None
            return
        self.__starttime = time
        self.__sdate = '{}{:02d}{:02d}'.format(time.year,
                                               time.month,
                                               time.day)

    def get_endtime(self):
        return self.__endtime

    def set_endtime(self, time):
        if time is None:
            self.__endtime = None
            self.__edate = None
            return
        self.__endtime = time
        self.__edate = '{}{:02d}{:02d}'.format(time.year,
                                               time.month,
                                               time.day)

    def get_volcano(self):
        return self.__volcano

    def set_volcano(self, value):
        self.__volcano = value

    def get_site(self):
        return self.__site

    def set_site(self, value):
        self.__site = value

    def get_channel(self):
        return self.__channel

    def set_channel(self, value):
        self.__channel = value

    @property
    def sitedir(self):
        try:
            return os.path.join(self.rootdir,
                                self.volcano,
                                self.site,
                                self.channel)
        except TypeError:
            return None

    starttime = property(get_starttime, set_starttime)
    endtime = property(get_endtime, set_endtime)
    volcano = property(get_volcano, set_volcano)
    site = property(get_site, set_site)
    channel = property(get_channel, set_channel)


class VolcanoMetadata:
    """
    Read volcano/station/channel metadata
    """

    def __init__(self, configfile):
        try:
            with open(configfile, 'r') as fh:
                cfg = yaml.safe_load(fh)
        except OSError:
            cfg = yaml.safe_load(configfile)
        self.data = cfg['metadata']

    def get_available_volcanoes(self):
        return [vol['name'] for vol in self.data['volcano']]

    def get_available_streams(self, volcano):
        """
        Get available sensor streams for a given volcano.

        Parameters
        ----------
        volcano : str
            Volcano name 

        Returns
        -------
        list 
            List of streams in the format 'NET.STA.LOC.CHAN' 
        """
        _streams = [] 
        for _v in self.data['volcano']:
            if _v['name'].lower() == volcano.lower():
                for _n in _v['network']:
                    net_code = _n['net_code']
                    for _s in _n['stations']:
                        sta_code = _s['sta_code']
                        location = _s['location']
                        for _c in _s['channel']:
                            chan_code = _c['code']
                            _stream = '.'.join((net_code, sta_code, location, chan_code))
                            _streams.append(_stream)
        return _streams

    def get_site_information(self, sitename):
        """
        Get site information for a given site. Note 
        that if the same site name exists at multiple
        volcanoes, the first match will be returned.

        Parameters
        ----------
        sitename : str 
            Three-letter site code. 

        Returns
        -------
        dict 
            Dictionary with site information as follows:
            {'volcano': str, 'site': str, 'channels': list, 
             'latitude': float, 'longitude': float, 'starttime': datetime}            
        """
        for _v in self.data['volcano']:
            for _n in _v['network']:
                for _s in _n['stations']:
                    if _s['sta_code'] == sitename:
                        try:
                            starttime = datetime.strptime(_s['starttime'],
                                                        "%Y-%m-%dT%H:%M:%SZ")
                        except KeyError:
                            starttime = None

                        try:
                            channels = [c['code'] for c in _s['channel']]
                        except KeyError:
                            channels = []
                        return dict(volcano=_v['name'],
                                    site=_s['sta_code'],
                                    channels=channels,
                                    latitude=float(_s['latitude']),
                                    longitude=float(_s['longitude']),
                                    starttime=starttime)
        return None

    def get_eruption_dates(self, volcano):
        """
        Get eruption dates for a given volcano.

        Parameters
        ----------
        volcano : str 
            Volcano name. 

        Returns
        -------
        list 
            List of datetime.date objects representing eruption dates. 
        """
        eruption_dates = []
        metadata = [v for v in self.data['volcano']
                    if v['name'] == volcano]
        er = metadata[0]["eruptions"]
        if er:
            for e in er:
                d, m, y = e.split('-')
                eruption_dates.append(date(int(y), int(m), int(d)))
        return eruption_dates

    def get_unrest_periods(self, volcano):
        """
        Get unrest periods for a given volcano.

        Parameters
        ----------
        volcano : str 
            Volcano name. 

        Returns
        -------
        list 
            List of datetime.date objects representing start and end dates. 
        """
        unrest_dates = []
        metadata = [v for v in self.data['volcano']
                    if v['name'] == volcano]
        ur = metadata[0]["unrest periods"]
        if ur:
            for u in ur:
                ds, ms, ys = u['starttime'].split('-')
                de, me, ye = u['endtime'].split('-')
                unrest_dates.append([date(int(ys), int(ms), int(ds)),
                                     date(int(ye), int(me), int(de))])
        return unrest_dates


if __name__ == '__main__':
    import doctest
    doctest.testmod()



import datetime
import os
import shutil
import tempfile
import unittest
import numpy as np

import h5py
from obspy import Trace, UTCDateTime
from obspy.signal.util import enframe
import xarray as xr

from zizou.util import (stride_windows,
                       demean,
                       apply_hanning,
                       window_array,
                       trace_window_data,
                       trace_window_times,
                       apply_freq_filter,
                       round_time,
                       xarray2hdf5,
                       test_signal)
from zizou.ssam import SSAM
from zizou.rsam import RSAM


class UtilTestCase(unittest.TestCase):

    def setUp(self):
        self.tempdir = tempfile.mkdtemp()

    def tearDown(self):
        if os.path.isdir(self.tempdir):
            shutil.rmtree(self.tempdir)

    def test_stride_windows(self):
        data = np.random.randint(0, 10, size=10)
        test_array = np.vstack((data[0:5], data[3:8]))
        ds = stride_windows(data, 5, 2)
        np.testing.assert_array_equal(test_array.T, ds)
        ds = stride_windows(data[0:8], 5, 2)
        np.testing.assert_array_equal(test_array.T, ds)
        # Because the returned array is just a
        # different view on the original array,
        # you can't assign to it
        with self.assertRaises(ValueError):
            ds[0, 0] = 5

        # But because ds is just a view, changing the
        # original array will change ds is, maybe,
        # unexpected ways:
        data[3] = 100
        self.assertEqual(ds[0, 1], 100)
        
    def test_obspy_enframe(self):
        """
        Compare stride_windows to obspy's enframe which doesn't
        use 'as_strided'.
        """
        data = np.random.randint(0, 10, size=10)
        ds = window_array(data, 4, 2, taper=False, padval=None)
        ds1, nwin, nowin  = enframe(data, np.ones(4), 2)
        np.testing.assert_array_equal(ds.T, ds1)

    def test_demean(self):
        a = np.arange(9.)
        ds = stride_windows(a, 3, 2)
        np.testing.assert_array_equal(demean(ds, axis=0)[0], -np.ones(7))
        a[1] = np.nan
        np.testing.assert_array_equal(demean(ds, axis=0)[1, 1], -0.5)

    def test_taper(self):
        a = np.arange(9)
        ds = stride_windows(a, 3, 1)
        dsw, wvals = apply_hanning(ds, return_window=True)
        np.testing.assert_array_equal(dsw[:, 0], np.array([0, 1., 0]))
        np.testing.assert_array_equal(dsw[1], ds[1])
        np.testing.assert_array_equal(wvals,
                                      np.array([0, 1, 0])[:, np.newaxis])

    def test_window_array(self):
        data = np.arange(10)
        # This will produce 4 windows with 5 elements each
        # overlapping by 3 elements; the last window will
        # contain one 0 to bring it to size 5
        res = window_array(data, 5, 3, remove_mean=False,
                           taper=False, padval=0.)
        self.assertEqual(res.shape, (5, 4))
        self.assertEqual(res[-1, -1], 0)
        # The returned array is just a different view of the
        # original so you can't assign to it
        with self.assertRaises(ValueError):
            res[0, 0] = 100
        # Same as above but also remove the mean from each
        # column
        resm = window_array(data, 5, 3, remove_mean=True,
                            taper=False, padval=0.)
        np.testing.assert_array_equal(resm.sum(axis=0),
                                      np.zeros(resm.shape[1]))
        # The returned array is a new copy of the original
        # so you now can assign to it
        resm[0, 0] = 100
        self.assertEqual(resm[0, 0], 100)

        # Now remove the mean and also taper each window. This
        # is the default behaviour
        rest, wvals = window_array(data, 5, 3, padval=0.)
        np.testing.assert_array_equal(rest[1], resm[1] * wvals[1])

        # This also works if we pad with np.nan
        rest, wvals = window_array(data, 5, 3, padval=np.nan)
        res_cp = res.copy()
        res_cp[-1, -1] = np.nan
        res_test = (res_cp[:4] - res_cp[:4, -1].sum()/4) * wvals[:4]
        np.testing.assert_array_equal(rest[:4, -1], res_test[:4, -1])

    def test_trace_window_data(self):
        tr = _dummy_trace_data(901)
        interval = 1.
        result = list(trace_window_data(tr, interval))
        self.assertEqual(len(result), 9)
        np.testing.assert_array_equal(tr.data[100:201], result[1].data)
        with self.assertRaises(ValueError):
            list(trace_window_times(tr, 100000))

    def test_trace_window_times(self):
        n_samp = 901
        hz = 100.
        tr = _dummy_trace_data(n_samp, hz)
        tr_len = (n_samp - 1) / hz

        # test simple case
        interval = 1.
        n_windows = int(tr_len / interval)
        result = list(trace_window_times(tr, interval, interval))
        self.assertEqual(len(result), n_windows)
        self.assertEqual(tr.stats.starttime + 3 * interval, result[3][0])
        # test cases where final window len < interval
        interval = 1.05
        n_windows = int(tr_len / interval)
        result = list(trace_window_times(tr, interval, 1))
        self.assertEqual(len(result), n_windows)
        result = list(trace_window_times(tr, interval, 0.2))
        self.assertEqual(len(result), n_windows + 1)
        with self.assertRaises(ValueError):
            list(trace_window_times(tr, 100000))

    def test_apply_freq_filter(self):
        f = (2, 10)
        ftype = "bp"
        tr_untouched = _dummy_trace_data()
        tr1 = tr_untouched.copy()
        tr2 = tr_untouched.copy()
        apply_freq_filter(tr1, ftype, f)
        tr2.filter(
            "bandpass", freqmin=f[0], freqmax=f[1], corners=4, zerophase=False
        )
        np.testing.assert_array_equal(tr1.data, tr2.data)

    def test_round_time(self):
        """
        Test find closest time to interval.
        """
        time = UTCDateTime(2021,5,21,13,8)
        self.assertEqual(round_time(time, 600),
                         UTCDateTime(2021,5,21,13,10))

    def test_xarray2hdf5(self):
        """
        Test writing xarray data to hdf5.
        """
        linfreqs = np.arange(0, 25.1, 0.1)
        s = SSAM(interval=60, per_lap=.9, smooth=10,
                 frequencies=linfreqs)
        td = tempfile.mkdtemp()

        xdf = s.compute(test_signal(sinusoid=False))
        xarray2hdf5(xdf, td)
 
        xdf_test = xr.open_dataset(os.path.join(td, 'ssam.nc'), group="original")
        np.testing.assert_array_equal(xdf['ssam'].values,
                                      xdf_test['ssam'].values)
        np.testing.assert_array_equal(xdf['frequency'].values,
                                      np.squeeze(xdf_test['frequency'].values))
        # minor differences can occur on the level of nanoseconds; ensure
        # differences are less than 1 microsecond
        dt = np.abs((xdf_test['datetime'].values - xdf['datetime'].values)).max()
        self.assertTrue(dt < np.timedelta64(1, 'us'))
        

    def test_xarray2hdf5_with_gaps(self):
        """
        Test writing xarray data to hdf5 with gaps.
        """
        linfreqs = np.arange(0, 25.1, 0.1)
        r = RSAM()

        xdf1 = r.compute(test_signal(nsec=3600, starttime=UTCDateTime(2022,7,18,8,0,0)))
        xdf2 = r.compute(test_signal(nsec=3600, starttime=UTCDateTime(2022,7,18,12,0,0)))

        xarray2hdf5(xdf1, self.tempdir)
        xarray2hdf5(xdf2, self.tempdir)
        xdf_test = xr.open_dataset(os.path.join(self.tempdir, 'rsam.nc'),
                                   group='original')
        self.assertEqual(xdf_test.rsam.isnull().sum(), 18)
         
    @unittest.expectedFailure
    def test_xarray2hdf5_multi_access(self):
        """
        Test writing xarray data to hdf5 while the file is open. This is currently
        not working with NetCDF4. See the following discussions for reference:
        https://github.com/pydata/xarray/issues/2887
        https://stackoverflow.com/questions/49701623/is-there-a-way-to-release-the-file-lock-for-a-xarray-dataset
        """
        linfreqs = np.arange(0, 25.1, 0.1)
        r = RSAM()

        xdf1 = r.compute(test_signal(nsec=3600, starttime=UTCDateTime(2022,7,18,8,0,0)))
        xdf2 = r.compute(test_signal(nsec=3600, starttime=UTCDateTime(2022,7,18,12,0,0)))
        tf = os.path.join(self.tempdir, tempfile.mktemp())

        xarray2hdf5(xdf1, tf)
        # f = h5py.File(tf, 'r')
        xdf_dummy = xr.open_dataset(tf, group='original', engine='h5netcdf')
        with self.assertRaises(OSError):
            xarray2hdf5(xdf2, tf)
        xdf_test = xr.open_dataset(tf, group='original', engine='h5netcdf')
        self.assertEqual(xdf_test.rsam.isnull().sum(), 18)
 

def _dummy_trace_data(n=2048, hz=100.):
    t = np.linspace(0, 40 * np.pi, n) + np.linspace(0, 13 * np.pi, n)
    return Trace(
        np.sin(t),
        header={"starttime": "2020-01-01T12:00:00", "sampling_rate": hz}
    )


if __name__ == '__main__':
    unittest.main()



"""
The main PCA script
"""
import logging
import os
import pickle
from argparse import ArgumentParser
from configparser import ConfigParser

from obspy import UTCDateTime
import sklearn.impute
import sklearn.decomposition
import xarray as xr

from zizou import AnomalyDetectionBaseClass, get_data
from zizou.data import FeatureRequest, VolcanoMetadata


logger = logging.getLogger(__name__)


class ModelException(Exception):
    pass


class PCA(AnomalyDetectionBaseClass):

    def __init__(self, modelfile='pcaModel.pkl', pca_features=None, interval=600,
                 n_components=4):
        self.pca_features = pca_features
        self.modelfile = modelfile
        self.interval = interval
        self.n_components = n_components
        self.__pca_model = None
        self.__mean = None
        self.__std = None
        self.__feature = None
        self.__training_run = False

    @property
    def pca_model(self):
        return self.__pca_model

    @property
    def mean(self):
        return self.__mean

    @property
    def std(self):
        return self.__std

    @property
    def feature(self):
        return self.__feature

    def fit(self, data):
        if os.path.isfile(self.modelfile):
            self._read_model_parameters(self.modelfile)
        else:
            pca = sklearn.decomposition.PCA(n_components=self.n_components)
            data = self.get_features(self.fq).values
            pca.fit(data)
            self.__pca_model = pca
            self.__mean = mns
            self.__std = stds
            self._write_model_parameters(self.modelfile)
        self.__training_run = True

    def _write_model_parameters(self, modelfile):
        dirName = os.path.dirname(modelfile)
        if not os.path.exists(dirName):
            os.makedirs(dirName)
        with open(modelfile, 'wb') as fh:
            pickle.dump([self.__pca_model, self.__mean, self.__std], fh)

    def _read_model_parameters(self, modelfile):
        with open(modelfile, 'rb') as fh:
            self.__pca_model, self.__mean, self.__std = pickle.load(fh)

    def infer(self, fq, savedir=None, overwrite=False):
        '''
        :param fq: Object to request features.
        :type fq: :class:`FeatureRequest`
        :param savedir:
        :param overwrite:
        :return:
        '''
        if not self.__training_run:
            msg = "Run 'training' first."
            raise ModelException(msg)

        feats, mns, stds = self.featureStack(fq)
        vals = self.pca_model.transform(feats.values.T)
        keys = ['pc%d' % i for i in range(self.n_components)]
        foo = xr.DataArray(vals, coords=[feats.datetime.values,
                                         keys],
                           dims=['datetime', 'pca_component'])
        xdf = xr.Dataset({'pca': foo})
        xdf.attrs['starttime'] = fq.starttime.isoformat()
        xdf.attrs['endtime'] = fq.endtime.isoformat()
        xdf.attrs['station'] = fq.site
        if savedir is not None:
            self.save_model_output(xdf, savedir, overwrite=overwrite)
        return xdf


def runPca(argv=None):
    parser = ArgumentParser(prog='compute_pca',
                            description=__doc__.strip())
    parser.add_argument('-c', '--config', type=str,
                        default=get_data('package_data/config.ini'),
                        help='Path to config file.')
    parser.add_argument('-m', '--metadata', type=str,
                        default=get_data('package_data/metadata.json'),
                        help='Volcano metadata json')
    parser.add_argument('-o', '--out_dir', type=str,
                        default='/tmp',
                        help='Output dir')
    parser.add_argument('-s', '--starttime', type=str,
                        help='Start of the raw data window')
    parser.add_argument('-e', '--endtime', type=str,
                        help='End of the raw data window')
    parser.add_argument('--overwrite', action='store_true',
                        help='Overwrite existing files instead merging.')

    args = parser.parse_args(argv)
    endtime = UTCDateTime.now()
    starttime = endtime-3600.
    if args.starttime:
        starttime = UTCDateTime(args.starttime)
    if args.endtime:
        endtime = UTCDateTime(args.endtime)

    runPcaCore(args.config, args.metadata, args.out_dir,
               endtime.datetime, starttime.datetime,
               overwrite=args.overwrite)


def runPcaCore(config_file, metadata_file, out_dir,
               end_time=UTCDateTime.now(),
               start_time=None, overwrite=False):
    config = ConfigParser()
    config.read(config_file)
    volcanoes = config['DEFAULT']['volcanoes'].split(',')
    # List of features to use in PCA
    pca_features = config['PCA']['features'].split(',')
    vm = VolcanoMetadata(file=metadata_file)
    for volcano in volcanoes:
        vm.set_volcano_name(volcano)
        streams = vm.get_seismic_network_streams()
        for stream in streams:
            net, site, loc, comp = stream[0].split('.')
            st_stream = UTCDateTime(stream[1])
            # Check that input dates are valid for channel
            if start_time is None:
                if end_time > st_stream:
                    start_time = UTCDateTime(st_stream.year,
                                             st_stream.month,
                                             st_stream.day+1, 0, 0, 0)
            fq = FeatureRequest(volcano=volcano, site=site, channel=comp,
                                starttime=start_time, endtime=end_time)
            msg = "Compute PCA for: {}, {}, {}, {}, {}"
            msg = msg.format(volcano, site, comp, start_time, end_time)
            logging.info(msg)
            savedir = os.path.join(out_dir, volcano, site, comp, 'PCA')
            modelfile = os.path.join(savedir, "pca.pkl")
            pca = PCA(modelfile=modelfile, pca_features=pca_features)
            pca.fit(fq)
            pca.infer(fq, savedir=savedir, overwrite=overwrite)

if __name__ == '__main__':
    runPca()



import os
import shutil
import tempfile
import unittest
import warnings

import numpy as np
from obspy import UTCDateTime
import xarray as xr

from zizou.spectral_features import SpectralFeatures
from zizou.util import test_signal, xarray2hdf5


class SpectralTestCase(unittest.TestCase):

    def setUp(self):
        self.outdir = tempfile.mkdtemp()

    def tearDown(self):
        if os.path.isdir(self.outdir):
            shutil.rmtree(self.outdir)

    sup = np.testing.suppress_warnings()
    sup.filter(module=np.matrixlib.defmatrix)  # module must match exactly

    def test_spectral(self):
        sf = SpectralFeatures(filtertype='bandpass', filterfreq=(0.5, 25))
        test_trace = test_signal(addchirp=False, frequencies=[5.],
                                 amplitudes=[1.0], phases=[0.0],
                                 sinusoid=False, noise=False)
        xds = sf.compute(test_trace)
        self.assertAlmostEqual(xds.predom_freq.data[0], 5, 0)
        self.assertAlmostEqual(xds.central_freq.data[0], 5, 0)
        self.assertAlmostEqual(xds.bandwidth.data[0], .062, 3)

    def test_spectral_with_gaps(self):
        sf = SpectralFeatures(filtertype='bandpass', filterfreq=(0.5, 25))
        test_trace = test_signal(addchirp=False, frequencies=[5.],
                                 amplitudes=[1.0], phases=[0.0],
                                 sinusoid=False, noise=False, gaps=True)
        xds = sf.compute(test_trace)
        self.assertAlmostEqual(xds.predom_freq.data[0], 5, 0)
        self.assertAlmostEqual(xds.central_freq.data[0], 5, 0)
        self.assertAlmostEqual(xds.bandwidth.data[0], .062, 3)
        self.assertEqual(np.isnan(xds.predom_freq.values).sum(), 3)

    def test_realtime(self):
        """
        Test with 1 hour windows as in the real-time computation
        """
        sf = SpectralFeatures(filtertype='highpass', filterfreq=(0.5, None))
        starttime = UTCDateTime(2020, 4, 21, 0, 0, 0)
        nsec = 86500 # Slightly longer than 24 h to avoid rounding issues
        step = 3600
        tr = test_signal(nsec=nsec, starttime=starttime, gaps=True)

        startwin = starttime
        while True:
            endwin = startwin + step
            if endwin > tr.stats.endtime:
                break
            ntr = tr.slice(startwin, endwin)
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                xdf = sf.compute(ntr)
            for featureName in list(xdf.data_vars.keys()):
                xarray2hdf5(xdf, self.outdir)
            startwin = endwin
        for feature in ['bandwidth', 'central_freq', 'predom_freq']:
            xdf = xr.load_dataset(os.path.join(self.outdir, feature+'.nc'),
                                  group='original')
            self.assertEqual(xdf[feature].data.shape, (144,))





import numpy as np
import obspy.signal.freqattributes
import obspy.signal.hoctavbands
import obspy.signal.util
import pandas as pd
import scipy.fftpack
import scipy.integrate
import yaml

from zizou import FeatureBaseClass
from zizou.util import apply_freq_filter, trace_window_data


class SpectralFeatures(FeatureBaseClass):
    
    features = ['central_freq', 'bandwidth', 'predom_freq']

    def __init__(self, interval=600., filtertype='highpass',
                 filterfreq=(0.5, None), ncep=8,
                 configfile=None):
        super(SpectralFeatures, self).__init__()
        self.interval = float(interval)
        self.filtertype = filtertype
        self.filterfreq = filterfreq
        self.ncep = int(ncep)
        if configfile is not None:
            try:
                with open(configfile, 'r') as fh:
                    c = yaml.safe_load(fh)
            except OSError:
                c = yaml.safe_load(configfile)
            self.interval = c['default'].get('interval', interval)
            cs = c.get('spectral_features')
            if cs is not None:
                self.filtertype = cs.get('filtertype', filtertype) 
                freqs = cs.get('filterfreq')
                if freqs is not None:
                    self.filterfreq = (freqs['low'], freqs['high'])
                self.ncep = cs.get('num_cepstral', ncep)

    def compute(self, trace):
        """
        :param trace: The seismic waveform data
        :type trace: :class:`obspy.Trace`
        :return:
        DataFrame containing
        - central frequency (acceleration) in Hz
        - bandwidth parameter (acceleration)
        - predominant frequency (acceleration) in Hz
        - eight cepstral coefficients
        """
        if len(trace) < 1:
            msg = "Trace is empty."
            raise ValueError(msg)
        fs = trace.stats.sampling_rate

        keys = ['central_freq', 'bandwidth', 'predom_freq']
        # keys += [f'cepstral_coef_{ind}' for ind in np.arange(self.ncep)]

        # initialise dataframe
        feature_data = []
        feature_idx = []

        # loop through data in interval blocks
        for tr_int in trace_window_data(trace, self.interval, min_len=0.8):
            if np.any(np.isnan(tr_int.data)):
                vals = np.full(len(keys), np.nan)
            else:
                tr_int.detrend('constant')
                tr_int.detrend('linear')
                apply_freq_filter(tr_int, self.filtertype, self.filterfreq)
                tr_int_diff = tr_int.differentiate()

                # Calculate psdf
                nfft = obspy.signal.util.next_pow_2(tr_int_diff.stats.npts)
                frequency = np.linspace(0, fs, nfft + 1)
                frequency = frequency[0:nfft // 2]
                data_f = scipy.fftpack.fft(tr_int_diff.data, nfft)
                data_psd = np.abs(data_f[0:nfft // 2]) ** 2

                # Interpolate onto logspace frequency vector
                f_ls = np.logspace(-2, 2, 401)  # These constants should be accessible
                data_psd_raw = np.interp(f_ls, frequency, data_psd)
                data_psd_ls = obspy.signal.util.smooth(data_psd_raw, 3)

                central_freq = self.centralFrequency(data_psd_ls, f_ls)
                bandwidth = self.bandwidth(data_psd_ls, f_ls)
                predom_freq = self.predominantFrequency(data_psd_ls.data, f_ls)
                # cepstral_coef = self.cepstralCoefficients(
                #     tr_int.data, fs, self.ncep
                # )
                vals = [central_freq, bandwidth, predom_freq]
                #vals = np.concatenate((vals, cepstral_coef, hob), axis=None)

            feature_data.append(vals)
            feature_idx.append(
                tr_int.stats.starttime.strftime('%Y-%m-%dT%H:%M:%S')
            )

        xdf = pd.DataFrame(
            data=feature_data,
            index=pd.DatetimeIndex(feature_idx, name="datetime"),
            columns=keys,
            dtype=float,
        ).to_xarray()
        xdf.attrs['starttime'] = trace.stats.starttime.isoformat()
        xdf.attrs['endtime'] = trace.stats.endtime.isoformat()
        xdf.attrs['station'] = trace.stats.station

        self.feature = xdf
        self.trace = trace
        return self.feature

    @staticmethod
    def predominantFrequency(psd, freq):
        """
        The peak frequency of the acceleration power spectral
        density function, in Hz.
        """
        predom_freq = freq[np.argmax(psd)]
        return predom_freq

    @staticmethod
    def centralFrequency(psd, freq):
        """
        The 'central' frequency of the acceleration power
        spectral density function (Hz). Defined as the
        square root of the ratio between the second and
        zero-th spectral moments. The spectral moments
        are calculated with log-spaced frequencies.
        """
        m0 = 2 * scipy.integrate.simps(freq ** 0 * psd)
        m2 = 2 * scipy.integrate.simps(freq ** 2 * psd)
        return np.sqrt(m2 / m0)

    @staticmethod
    def bandwidth(psd, freq):
        """
        The Vanmarcke (1970) bandwidth parameter, q where
        0 <= q <= 1. Calculated using the spectral moments
        of the acceleration power spectral density function.
        0 is a Dirac delta PSDF (sine wave), 1 is a flat PSDF
        (white signal).
        """
        m0 = 2 * scipy.integrate.simps(freq**0 * psd)
        m1 = 2 * scipy.integrate.simps(freq**1 * psd)
        m2 = 2 * scipy.integrate.simps(freq**2 * psd)
        return np.sqrt(1 - m1**2 / m2 / m0)

    @staticmethod
    def cepstralCoefficients(data, fs, ncep=8, nfilters=10, window="Hamming"):
        cep = obspy.signal.freqattributes.log_cepstrum(
            data=np.expand_dims(data, axis=0),
            fs=fs,
            nc=ncep,
            p=nfilters,
            n=None,
            w=window
        )
        return cep



import shutil
import tempfile
import unittest

import numpy as np
import os

import pandas as pd
import xarray as xr

from zizou import FeatureBaseClass

dt1 = '2020-01-03 00:10:00'
dt2 = '2020-01-03 00:20:00'
dt3 = '2020-01-03 00:30:00'
dt4 = '2020-01-03 00:40:00'
dt5 = '2020-01-04 00:10:00'
dt6 = '2020-01-04 00:30:00'
dt7 = '2020-01-04 00:50:00'


class BaseClassTestCase(unittest.TestCase):

    def setUp(self):
        self.savedir = tempfile.mkdtemp() 

    def tearDown(self):
        if os.path.isdir(self.savedir):
            shutil.rmtree(self.savedir)

    def test_exceptions(self):
        """
        Test that the right exceptions are raised.
        """
        bc = FeatureBaseClass()
        with self.assertRaises(NotImplementedError):
            bc.compute()

        with self.assertRaises(TypeError):
            bc.save()

    def test_append(self):
        """
        Test appending existing HDF5 files.
        """
        featureName = '1DFeature'
        xdf1 = self.make1DDataSet(dt1, dt3, [1, 2], featureName)
        bc1 = FeatureBaseClass()
        bc1.feature = xdf1
        bc1.save(self.savedir) #here we overwrite the file first to make sure it's created from scratch

        testfile = os.path.join(self.savedir, featureName+'.nc')
        with xr.load_dataset(testfile, group='original') as xds1:
            self.assertEqual(np.datetime64(xds1.attrs['starttime']).astype('datetime64[ms]'),
                             np.datetime64(dt1).astype('datetime64[ms]'))
            self.assertEqual(np.datetime64(xds1.attrs['endtime']).astype('datetime64[ms]'),
                             np.datetime64(dt3).astype('datetime64[ms]'))

        xdf2 = self.make1DDataSet(dt3, dt4, [3, 4], featureName)
        bc2 = FeatureBaseClass()
        bc2.feature = xdf2
        bc2.save(self.savedir)

        with xr.load_dataset(testfile, group='original') as xds1:
            self.assertEqual(np.datetime64(xds1.attrs['starttime']).astype('datetime64[ms]'),
                             np.datetime64(dt1).astype('datetime64[ms]'))
            self.assertEqual(np.datetime64(xds1.attrs['endtime']).astype('datetime64[ms]'),
                             np.datetime64(dt4).astype('datetime64[ms]'))
            np.testing.assert_array_almost_equal(xds1['1DFeature'].values,
                                                 np.array([1., 2., 3., 4.]), 5)

    def test_2D_append(self):
        """
        Test appending when the feature is an xarray Dataset with 2D
        data.
        """
        xdf1, xdf2 = self.make2DDataSets()

        bc1 = FeatureBaseClass()
        bc1.feature = xdf1
        bc1.save(self.savedir)

        bc2 = FeatureBaseClass()
        bc2.feature = xdf2
        bc2.save(self.savedir)

        testfile = os.path.join(self.savedir, '2Dfeature.nc')
        with xr.open_dataset(testfile, group='original') as xds:
            self.assertEqual(np.datetime64(xds.attrs['starttime']).astype('datetime64[ms]'),
                             np.datetime64(dt5).astype('datetime64[ms]'))
            self.assertEqual(np.datetime64(xds.attrs['endtime']).astype('datetime64[ms]'),
                             np.datetime64(dt7).astype('datetime64[ms]'))

            np.testing.assert_array_almost_equal(xds['2Dfeature'].values,
                                                 np.array([[1., 2., 6., 7., 8.],
                                                           [4., 5., 9., 10., 11.]]))

    def make1DDataSet(self, dt1, dt2, values, featureName):
        t = pd.date_range(dt1, dt2,periods=2)
        xdf = xr.DataArray(values, coords=dict(datetime=t), dims=['datetime'])
        xdf = xr.Dataset({featureName: xdf})
        xdf.attrs['station'] = 'WIZ'
        return xdf

    def make2DDataSets(self):
        t1 = pd.date_range(dt5, dt6, freq='10min')
        values = np.arange(1, 7).reshape((2, 3))
        foo = xr.DataArray(values,
                           coords=[np.array([10, 20]), t1],
                           dims=['frequency', 'datetime'])
        xdf1 = xr.Dataset({'2Dfeature': foo})
        xdf1.attrs['starttime'] = t1[0].isoformat()
        xdf1.attrs['endtime'] = t1[-1].isoformat()
        xdf1.attrs['station'] = 'WIZ'

        # Note that t1 and t2 overlap so the final result should have the
        # first row of the second dataset overwritten by the last row of
        # the first dataset
        t2 = pd.date_range(dt6, dt7, freq='10min')
        values = np.arange(6, 12).reshape((2, 3))
        bar = xr.DataArray(values,
                           coords=[np.array([10, 20]), t2],
                           dims=['frequency', 'datetime'])

        xdf2 = xr.Dataset({'2Dfeature': bar})
        xdf2.attrs['starttime'] = t2[0].isoformat()
        xdf2.attrs['endtime'] = t2[-1].isoformat()
        xdf2.attrs['station'] = 'WIZ'

        return xdf1, xdf2


if __name__ == '__main__':
    unittest.main()



"""
Compute the Real-time Seismic-Amplitude Measurement (RSAM).
"""
import numpy as np
from obspy import UTCDateTime
import pandas as pd
import scipy.integrate
import xarray as xr

from zizou import FeatureBaseClass
from zizou.util import (apply_freq_filter,
                       trace_window_data,
                       trace_window_times,
                       round_time)
import yaml

__all__ = ["RSAM", "EnergyExplainedByRSAM"]


def rsam(arr, axis=None):
    return np.absolute(arr).mean(axis=axis)


def rsam_wideband_energy_ratio(arr_rsam, arr_wideband, axis=-1):
    energy_rsam_band = scipy.integrate.simps(arr_rsam**2, axis=axis)
    energy_wide_band = scipy.integrate.simps(arr_wideband**2, axis=axis)
    return energy_rsam_band / energy_wide_band


class RSAM(FeatureBaseClass):
    """
    RSAM is the mean of the absolute value of a signal filtered
    between 2 and 5 Hz.

    :param interval: Length of the time interval in seconds over
                     which to compute RSAM.
    :type interval: int, optional
    :param filtertype: The type of filtering that is applied before
                       computing RSAM. Can be either band-pass ('bp'),
                       high-pass ('hp'), or low-pass ('lp').
    :type filtertype: str, optional
    :param filterfreq: The low and high cutoff frequency. If filtertype is 
                       highpass or lowpass, only the first or last value is
                       used, respectively.

    :type filterfreq: tuple, optional
    """

    features = ["rsam"]

    def __init__(self, interval=600., filtertype='bandpass',
                 filterfreq=(2, 5), configfile=None,
                 reindex=True):
        super(RSAM, self).__init__()
        self.interval = interval
        self.filtertype = filtertype
        self.filterfreq = filterfreq
        self.feature = None
        self.reindex = reindex

        if configfile is not None:
            try:
                with open(configfile, 'r') as fh:
                    c = yaml.safe_load(fh)
            except OSError:
                c = yaml.safe_load(configfile)
            self.interval = c['default'].get('interval', interval)
            cr = c.get('rsam')
            if cr is not None:
                self.filtertype = cr.get('filtertype', filtertype)
                freq = cr.get('filterfreq')
                if freq is not None:
                    self.filterfreq = (freq['low'], freq['high'])
                self.reindex = cr.get('reindex', reindex)

    def compute(self, trace):
        """
        :param trace: The seismic waveform data
        :type trace: :class:`obspy.Trace`
        """
        if len(trace) < 1:
            raise ValueError("Trace is empty.")
        tr = trace.copy()
        starttime = tr.stats.starttime
        endtime = tr.stats.endtime

        # handle gaps (i.e. NaNs)
        mdata = np.ma.masked_invalid(tr.data)
        tr.data = mdata
        st = tr.split()

        apply_freq_filter(st, self.filtertype, self.filterfreq)
        st.merge(fill_value=np.nan)
        st.trim(starttime, endtime, pad=True, fill_value=np.nan)
        tr = st[0]

        # initialise dataframe
        feature_data = []
        feature_idx = []

        # loop through data in interval blocks
        for trint in trace_window_data(tr, self.interval, min_len=0.8):
            absolute = np.absolute(trint.data)  # absolute value
            trint.data = absolute  # assign back to trace
            mean = trint.data.mean()
            # convert to nanometres so dealing with whole numbers
            feature_data.append(mean / 1e-9)
            feature_idx.append(
                trint.stats.starttime.strftime('%Y-%m-%dT%H:%M:%S')
            )

        xda = xr.DataArray(feature_data,
                           coords=[pd.DatetimeIndex(feature_idx)],
                           dims='datetime')
        xdf = xr.Dataset({'rsam': xda})
        xdf.attrs['starttime'] = trace.stats.starttime.isoformat()
        xdf.attrs['endtime'] = trace.stats.endtime.isoformat()
        xdf.attrs['station'] = trace.stats.station
        if self.reindex:
            starttime = UTCDateTime(str(xda.datetime.data[0]))
            starttime = round_time(starttime, self.interval)
            endtime = UTCDateTime(str(xda.datetime.data[-1]))
            endtime = round_time(endtime, self.interval)
            new_index = pd.date_range(starttime.datetime, endtime.datetime,
                                      freq="%dS" % int(self.interval))
            xdf = xdf.reindex(dict(datetime=new_index), method='nearest')
        self.feature = xdf
        return self.feature


class EnergyExplainedByRSAM(RSAM):
    """
    The proportion of signal energy in the RSAM
    bandwidth, relative to a 0.5-10 Hz bandwidth.
    """
    features = ["rsam_energy_prop"]

    def __init__(
        self,
        interval=600.,
        filtertype=None,
        filterfreq=(None, None),
        filtertype_wb="bandpass",
        filterfreq_wb=(0.5, 10.),
        configfile=None
    ):
        super(EnergyExplainedByRSAM, self).__init__(
            interval=interval,
            filtertype=filtertype,
            filterfreq=filterfreq,
            configfile=configfile,
        )
        self.filtertype_wb = filtertype_wb
        self.filterfreq_wb = filterfreq_wb

        # base params already read from configfile during superclass init
        # reopen and extract wideband filter params if any
        if configfile is not None:
            try:
                with open(configfile, 'r') as fh:
                    c = yaml.safe_load(fh)
            except OSError:
                c = yaml.safe_load(configfile)
            cr = c.get('rsam')
            if cr is not None:
                self.filtertype_wb = cr.get('filtertype_wb', filtertype_wb)
                freq = cr.get('filterfreq_wb')
                if freq is not None:
                    self.filterfreq_wb = (freq['low'], freq['high'])

    def compute(self, trace):
        """
        :param trace: The seismic waveform data
        :type trace: :class:`obspy.Trace`
        """
        if len(trace) < 1:
            msg = "Trace is empty."
            raise ValueError(msg)
        tr = trace.copy()
        starttime = tr.stats.starttime
        endtime = tr.stats.endtime

        # handle gaps (i.e. NaNs)
        mdata = np.ma.masked_invalid(tr.data)
        tr.data = mdata

        st_rsam = tr.split()
        st_wb = st_rsam.copy()

        apply_freq_filter(st_rsam, self.filtertype, self.filterfreq)
        st_rsam.merge(fill_value=np.nan)
        st_rsam.trim(starttime, endtime, pad=True, fill_value=np.nan)
        tr_rsam = st_rsam[0]

        # Get trace in wide bandwidth
        apply_freq_filter(st_wb, self.filtertype_wb, self.filterfreq_wb)
        st_wb.merge(fill_value=np.nan)
        st_wb.trim(starttime, endtime, pad=True, fill_value=np.nan)
        tr_wb = st_wb[0]

        # initialise dataframe
        feature_data = []
        feature_idx = []

        # loop through data in interval blocks
        for t0, t1 in trace_window_times(
            tr_rsam, self.interval, min_len=0.8
        ):
            energy_ratio = rsam_wideband_energy_ratio(
                tr_rsam.slice(t0, t1, nearest_sample=False).data,
                tr_wb.slice(t0, t1, nearest_sample=False).data
            )
            feature_data.append(energy_ratio)
            feature_idx.append(t0.strftime('%Y-%m-%dT%H:%M:%S'))

        xdf = pd.DataFrame(
            data=feature_data,
            index=pd.DatetimeIndex(feature_idx, name="datetime"),
            columns=['rsam_energy_prop'],
            dtype=float,
        ).to_xarray()

        xdf.attrs['starttime'] = trace.stats.starttime.isoformat()
        xdf.attrs['endtime'] = trace.stats.endtime.isoformat()
        xdf.attrs['station'] = trace.stats.station

        self.feature = xdf
        self.trace = tr
        return self.feature



import os
import shutil
import tempfile
import unittest
import warnings

import numpy as np
from obspy import Trace, UTCDateTime
from scipy.signal import find_peaks
import xarray as xr

from zizou.ssam import SSAM
from zizou.util import test_signal


class SSAMTestCase(unittest.TestCase):

    def setUp(self):
        np.random.seed(42)
        data = np.random.randint(0, 10, size=10)
        self.tr = Trace(data=data, header=dict(sampling_rate=1))
        self.outdir = tempfile.mkdtemp()

    def tearDown(self):
        if os.path.isdir(self.outdir):
            shutil.rmtree(self.outdir)

    def test_ssam(self):
        test_peaks = np.array([1., 3., 10.])
        s = SSAM(interval=60, per_lap=.9, smooth=10)
        spec = s.compute(test_signal(sinusoid=False, addchirp=False,
                                     frequencies=[1., 3., 10.]))
        peaks = find_peaks(spec['ssam'].data[:, 0], height=.1)[0]
        np.testing.assert_array_almost_equal(spec['frequency'].values[peaks],
                                             test_peaks, 2)
        
    def test_sonogram(self):
        s = SSAM(interval=60, per_lap=.9, smooth=10)
        test_trace = test_signal(offsets=[1000., 2000., 3000.],
                                 frequencies=[1., 2., 5.],
                                 amplitudes=[0.1, 1., 5.],
                                 sinusoid=False,
                                 addchirp=False)
        spec = s.compute(test_trace)
        idx1 = (int(np.abs(spec.datetime - 
                np.datetime64((test_trace.stats.starttime + 1500.)
                .datetime)).argmin()))
        idx2 = (int(np.abs(spec.datetime - 
                np.datetime64((test_trace.stats.starttime + 2600.)
                .datetime)).argmin()))
        idx3 = (int(np.abs(spec.datetime - 
                np.datetime64((test_trace.stats.starttime + 3500.)
                .datetime)).argmin()))
        self.assertAlmostEqual(float(spec.sonofrequency[spec['sonogram']
                                .isel(dict(datetime=idx1)).argmax()]), 2.295, 3)
        self.assertAlmostEqual(float(spec.sonofrequency[spec['sonogram']
                                .isel(dict(datetime=idx2)).argmax()]), 5.164, 3)
        self.assertAlmostEqual(float(spec.sonofrequency[spec['sonogram']
                                .isel(dict(datetime=idx3)).argmax()]), 11.62, 2)

    def test_filterbank(self):
        s = SSAM(interval=60, per_lap=.9, smooth=10)
        test_trace = test_signal(offsets=[1000., 2000., 3000.],
                                 frequencies=[0.1, 2., 5.],
                                 amplitudes=[0.1, 1., 5.],
                                 sinusoid=False,
                                 addchirp=False)
        spec = s.compute(test_trace)
        idx1 = (int(np.abs(spec.datetime - 
                np.datetime64((test_trace.stats.starttime + 1500.)
                .datetime)).argmin()))
        idx2 = (int(np.abs(spec.datetime - 
                np.datetime64((test_trace.stats.starttime + 2600.)
                .datetime)).argmin()))
        idx3 = (int(np.abs(spec.datetime - 
                np.datetime64((test_trace.stats.starttime + 3500.)
                .datetime)).argmin()))
        self.assertAlmostEqual(float(spec.fbfrequency[spec['filterbank']
                                .isel(dict(datetime=idx1)).argmax()]), 2.295, 3)
        self.assertAlmostEqual(float(spec.fbfrequency[spec['filterbank']
                                .isel(dict(datetime=idx2)).argmax()]), 5.164, 3)
        self.assertAlmostEqual(float(spec.fbfrequency[spec['filterbank']
                                .isel(dict(datetime=idx3)).argmax()]), 11.62, 2)


    def test_ssam_with_gaps(self):
        linfreqs = np.arange(0, 25.1, 0.1)
        test_peaks = np.array([0.1, 3., 5, 6.1, 10.])
        s = SSAM(interval=60, per_lap=.9, smooth=10,
                 frequencies=linfreqs)
        spec = s.compute(test_signal(gaps=True))
        peaks = find_peaks(spec['ssam'].data[:, 0], height=1e-1)[0]
        np.testing.assert_array_almost_equal(linfreqs[peaks], test_peaks, 2)

    def test_realtime(self):
        """
        Test with 1 hour windows as in the real-time computation
        """
        linfreqs = np.arange(0, 25.1, 0.1)
        testpeaks = np.array([0.1, 3., 5., 6., 10.])
        s = SSAM(interval=60, frequencies=linfreqs,
                 resample_int=(None, '10min'), timestamp='start')
        starttime = UTCDateTime(2020, 4, 21, 0, 0, 0)
        nsec = 86500 # Slightly longer than 24 h to avoid rounding issues
        step = 3600
        tr = test_signal(nsec=nsec, starttime=starttime, gaps=True)

        startwin = starttime
        while True:
            endwin = startwin + step
            if endwin > tr.stats.endtime:
                break
            ntr = tr.slice(startwin, endwin)
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                s.compute(ntr)
            s.save(self.outdir)
            startwin = endwin

        xdf = xr.open_dataset(os.path.join(self.outdir, 'ssam.nc'),
                              group='original')
        self.assertEqual(xdf.ssam.data.shape, (251, 144))
        peaks = find_peaks(xdf['ssam'].data[:, 0], height=1e-1)[0]
        np.testing.assert_array_almost_equal(linfreqs[peaks],
                                             testpeaks, 2)

    def test_filterbank(self):
        s = SSAM()
        _ = s.filterbank(10, 512, 100.)

    def test_config(self):
        """
        Test parsing of config parameters.
        """
        config = """
        default:
          interval: 600
        ssam:
          interval: 60
          per_lap: .9
          scale_by_freq: true
          timestamp: start
          frequencies:
            start: 0
            end: 25.1
            step: 0.1
          resample_int:
            upsample: null
            downsample: 10min
        """ 
        s = SSAM(configfile=config)
        self.assertEqual(s.resample_int, (None, '10min'))
        np.testing.assert_equal(s.frequencies, np.arange(0, 25.1, .1))





"""
Compute the displacement seismic amplitude ratio (DSAR).
"""
import numpy as np
from obspy import UTCDateTime
import pandas as pd
import xarray as xr

import yaml

from zizou import FeatureBaseClass
from zizou.util import trace_window_times, round_time


class DSAR(FeatureBaseClass):
    """
    DSAR is the ratio of
    low frequency to high frequency amplitudes, as suggested in:
    https://doi.org/10.1130/G46107.1

    The time-domain displacement signal is bandpass filtered in two
    frequency bands, 4.5 to 8 Hz and 8 to 16 Hz, and DSAR is defined
    as the median ratio of the absolute signals. The quantity is
    interpreted in the publication as a change in seismic attenuation
    in the vicinity of the seismic station.
    """
    features = ['dsar']

    def __init__(self, interval=600., filtertype='bandpass',
                 lowerfreqband=(4.5, 8),
                 higherfreqband=(8, 16),
                 configfile=None, reindex=True):
        super(DSAR, self).__init__()
        self.interval = float(interval)
        self.filtertype = filtertype
        self.lowerfreqband = lowerfreqband
        self.higherfreqband = higherfreqband
        self.reindex = reindex

        if configfile is not None:
            try:
                with open(configfile, 'r') as fh:
                    c = yaml.safe_load(fh)
            except OSError:
                c = yaml.safe_load(configfile)
            self.interval = c['default'].get('interval', interval)
            cd = c.get('dsar')
            if cd is not None:
                self.reindex = cd.get('reindex', reindex)
                self.filtertype = cd.get('filtertype', filtertype) 
                lfreq = cd.get('lowerfreqband')
                if lfreq is not None:
                    self.lowerfreqband = (lfreq['low'], lfreq['high']) 
                hfreq = cd.get('higherfreqband')
                if hfreq is not None:
                    self.higherfreqband = (hfreq['low'], hfreq['high']) 

    def compute(self, trace):
        """
        Compute the displacement seismic amplitude ratio (DSAR).

        # TODO:
        Publication is quite vague and doesn't have enough detail
        to understand the method. Unclear whether median is every
        10 minutes or every day. This affects the bootstrap calc-
        ulation too, but this might not be necessary..

        :param stream: The SLCN code for the seismic stream
        :type stream: str
        :param startdate: Datetime of the first sample.
        :type startdate: :class:`obspy.UTCDateTime`
        :param enddate: Datetime of the last sample.
        :type enddate: :class:`obspy.UTCDateTime`
        :param interval: Length of the time interval in seconds over
                         which to compute DSAR (default 10 mins)
        :type interval: int, optional
        """
        if len(trace) < 1:
            raise ValueError("Trace is empty.")

        tr_lf = trace.copy()
        # handle gaps (i.e. NaNs)
        mdata = np.ma.masked_invalid(tr_lf.data)
        tr_lf.data = mdata
        # Get second trace for alternative filtering
        tr_hf = tr_lf.copy()

        # Filter settings
        lf_f1, lf_f2 = self.lowerfreqband
        hf_f1, hf_f2 = self.higherfreqband
        st_lf = tr_lf.split()
        st_hf = tr_hf.split()
        # Integrate to displacement
        st_lf.integrate('cumtrapz')
        st_hf.integrate('cumtrapz')

        st_lf.filter(self.filtertype, freqmin=lf_f1, freqmax=lf_f2, corners=4,
                     zerophase=False)
        st_hf.filter(self.filtertype, freqmin=hf_f1, freqmax=hf_f2, corners=4,
                     zerophase=False)

        starttime = trace.stats.starttime
        endtime = trace.stats.endtime
        st_lf.merge(fill_value=np.nan)
        st_lf.trim(starttime, endtime, pad=True, fill_value=np.nan)
        st_hf.merge(fill_value=np.nan)
        st_hf.trim(starttime, endtime, pad=True, fill_value=np.nan)
        tr_lf = st_lf[0]
        tr_hf = st_hf[0]

        # DSAR calculation
        feature_data = []
        feature_idx = []

        # loop through data in interval blocks
        for t0, t1 in trace_window_times(tr_lf, self.interval, min_len=0.8):
            tr_lf_int = tr_lf.slice(t0, t1)
            tr_hf_int = tr_hf.slice(t0, t1)

            absolute_lf = np.absolute(tr_lf_int.data)
            absolute_hf = np.absolute(tr_hf_int.data)

            # calculate ratio
            if np.isnan(absolute_lf).any() or np.isnan(absolute_hf).any():
                median = np.nan
            else:
                m = absolute_hf > 0
                ratio = absolute_lf[m] / absolute_hf[m]
                median = np.median(ratio)

            feature_data.append(median)
            feature_idx.append(t0.strftime('%Y-%m-%dT%H:%M:%S'))
    
        xda = xr.DataArray(feature_data,
                           coords=[pd.DatetimeIndex(feature_idx)],
                           dims='datetime')
        xdf = xr.Dataset({'dsar': xda})
        xdf.attrs['starttime'] = trace.stats.starttime.isoformat()
        xdf.attrs['endtime'] = trace.stats.endtime.isoformat()
        xdf.attrs['station'] = trace.stats.station
        if self.reindex:
            starttime = UTCDateTime(str(xda.datetime.data[0]))
            starttime = round_time(starttime, self.interval)
            endtime = UTCDateTime(str(xda.datetime.data[-1]))
            endtime = round_time(endtime, self.interval)
            new_index = pd.date_range(starttime.datetime, endtime.datetime,
                                      freq="%dS" % int(self.interval))
            xdf = xdf.reindex(dict(datetime=new_index), method='nearest')
        self.feature = xdf
        return self.feature


import inspect
import os
import unittest

import numpy as np
from obspy import UTCDateTime
import pandas as pd

from zizou.dsar import DSAR
from zizou.data import DataSource


class DSARTestCase(unittest.TestCase):

    def setUp(self):
        self.data_dir = os.path.join(os.path.dirname(os.path.abspath(
            inspect.getfile(inspect.currentframe()))), "data")

    @unittest.skip("Tested code needs re-write as it's testing agains live FDSN server")
    def test_daily_ratio(self):
        dsar = DSAR(filtertype='bandpass', lowerfreqband=(4.5, 8),
                    higherfreqband=(8, 16))
        stream = 'KRVZ.10.EHZ.NZ'
        # dates are inclusive, and complete days
        startdate = UTCDateTime('2012-11-01')
        enddate = UTCDateTime('2012-11-03')
        ds = DataSource(sources=('fdsn',), fill_value='interpolate')
        site, loc, comp, net = stream.split('.')
        for tr in ds.get_waveforms(net, site, loc, comp,
                                   startdate, enddate):
            df = dsar.compute(tr)
        df_test = pd.read_csv(os.path.join(self.data_dir,
                                           'dsar_test.csv'),
                              index_col=0, parse_dates=True).to_xarray()
        np.testing.assert_array_almost_equal(df['dsar'].values, df_test['dsar'].values)

